"""
Dependency-Check report loader.
"""

import csv
import json
import logging
import re
from datetime import datetime
from os import path
from typing import List

from reportmix.config.property import ConfigProperty
from reportmix.loader import Loader
from reportmix.models import severity
from reportmix.models.issue import Issue
from reportmix.models.project import Project
from reportmix.models.subject import Subject
from reportmix.models.tool import Tool

# Configuration properties
PROPERTIES: List[ConfigProperty] = [
    ConfigProperty("report_file", "path to the report file", False, "dependency-check-report.csv")
]


class DependencyCheckLoader(Loader):
    """
    Dependency-Check report loader (CSV required, JSON optional).
    """

    def load(self) -> List[Issue]:
        """
        Load the Dependency Check report file (CSV required, JSON optional),
        parse it, map vulnerabilities to issues, and return the list.
        :return: List of vulnerabilities.
        """
        report_file_path = path.realpath(self.config["report_file"])
        if not (report_file_path.endswith(".csv") and path.exists(report_file_path)):
            logging.warning("Dependency check report ignored (file not found or not *.csv)")
            return []
        logging.debug("Loading report %s", report_file_path)
        try:
            # Load the JSON report to extract scan and project info
            json_report_file_path = re.sub(r"\.csv$", ".json", report_file_path)
            scan, project = {}, {}
            if path.exists(json_report_file_path):
                with open(json_report_file_path, "r", encoding="utf8") as json_report_file:
                    json_report = json.load(json_report_file)
                    scan = json_report["scanInfo"]
                    project = json_report["projectInfo"]

            # Load vulnerabilities from the CSV report and map them to issues
            with open(report_file_path, "r", newline='') as report_file:
                report = csv.DictReader(report_file, delimiter=',', quotechar='"')
                issues = []
                for row in report:
                    if "groupID" in project and "artifactID" in project:
                        project_identifier = project["groupID"] + ":" + project["artifactID"]
                    else:
                        project_identifier = row["Project"]
                    issues.append(Issue(
                        ref="",
                        identifier=row["CVE"],
                        name=row["CVE"],
                        type="VULNERABILITY",
                        category=row["CWE"],
                        description=row["Vulnerability"],
                        more="",
                        action="",
                        effort="",
                        # TODO Improve date parsing
                        analysis_date=datetime.strptime(row["ScanDate"][:24],
                                                        "%a, %d %b %Y %H:%M:%S"),
                        severity=severity.guess(row["CVSSv3_BaseSeverity"]),
                        score=row["CVSSv3"],
                        confidence=row["CPE Confidence"],
                        evidences=int(row["Evidence Count"]),
                        source=row["Source"],
                        source_date=None,
                        url="",
                        tool=Tool(
                            identifier="dependency_check",
                            name="Dependency-Check",
                            version=scan["engineVersion"] if "engineVersion" in scan else ""
                        ),
                        subject=Subject(
                            identifier=row["Identifiers"],
                            name=row["Description"],
                            description=row["DependencyName"],
                            version="",
                            location=row["DependencyPath"],
                            license=row["License"]
                        ),
                        project=Project(
                            identifier=project_identifier,
                            name=row["Project"],
                            version=project["version"] if "version" in project else ""
                        )
                    ))
                return issues
        except Exception as ex:
            logging.error("Failed to load, parse and map the report (%s)", ex)
            return []



"""
Severity model.
"""

from typing import List, Union


class Severity:
    """
    An issue severity.
    """

    def __init__(self, identifier: str, name: str):
        """
        Initialize an issue severity from a name and an id.
        :param identifier: Severity unique identifier
        :param name: Severity display name
        """
        self.identifier = identifier
        self.name = name

    def __str__(self) -> str:
        return self.identifier

    def __repr__(self) -> str:
        return "'" + self.identifier + "'"


#
# Constants
#

# Severities (based on CVSS v3 Ratings)
SEVERITIES: List[Severity] = [
    Severity("NOT_DEFINED", "Not Defined"),
    Severity("NONE", "None"),
    Severity("LOW", "Low"),
    Severity("MEDIUM", "Medium"),
    Severity("HIGH", "High"),
    Severity("CRITICAL", "Critical")
]


#
# Helpers
#

def guess(value: str) -> Union[Severity, None]:
    """
    Try to guess a severity from a given input value.
    :param value: Input value.
    :return: Guessed severity (None if guess failed)
    """
    val = value.strip() if value else value
    if not val:
        return SEVERITIES[0]
    # else
    val = val.lower()
    sev = None
    if val in ("info", "none"):
        sev = SEVERITIES[1]
    elif val in ("low", "minor"):
        sev = SEVERITIES[2]
    elif val in ("medium", "moderate"):
        sev = SEVERITIES[3]
    elif val in ("high", "major"):
        sev = SEVERITIES[4]
    elif val in ("critical", "blocker"):
        sev = SEVERITIES[5]
    return sev



"""
HTML report exporter.
"""

from collections import OrderedDict
from typing import List

import jinja2

from reportmix.exporter import Exporter
from reportmix.models.issue import Issue
from reportmix.models.severity import SEVERITIES


class HtmlExporter(Exporter):
    """
    Export a merged report to a HTML file.
    """

    def export(self, output_file: str, issues: List[Issue], fields: List[str]):
        # Load HTML template
        env = jinja2.Environment(
            loader=jinja2.PackageLoader("reportmix.exporters", "templates"),
            autoescape=jinja2.select_autoescape(['html'])
        )
        # Custom filters
        env.filters["limit"] = limit
        env.filters["prettyfield"] = pretty_field
        # Report template
        template = env.get_template('reportmix.html.jinja2')

        # Prepare templates values and statistics
        # Issues by tool
        tools = OrderedDict()
        tool_names = sorted({i.tool.name for i in issues},
                            key=lambda tool_name: tool_name.casefold())
        for tool in tool_names:
            tools[tool] = len([i for i in issues if i.tool.name == tool])
        # Issues by severity
        severities = OrderedDict()
        for severity in SEVERITIES:
            severities[severity.name] = len([i for i in issues if i.severity == severity])
        # Issues by type
        types = OrderedDict()
        type_names = sorted({i.type for i in issues}, key=lambda t: t.casefold())
        for tool in type_names:
            types[tool] = len([i for i in issues if i.type == tool])
        # Render and write report
        with open(output_file, "wb") as file:
            output = template.render(title=self.config["title"], logo=self.config["logo"],
                                     issues=[i.flatten() for i in issues], fields=fields,
                                     tools=tools, severities=severities, types=types)
            file.write(output.encode("utf-8"))


#
# Custom filters
#

def limit(value, max_length: int = 64) -> str:
    """
    Limit a value to a maximum length. If the value length is greater than max_length,
    the value is truncated, '...' is added at the end of the string,
    and the value is wrapped inside a <span> tag with a title containing
    the full string to allow the user to display it in a tooltip on hovering.
    :param value: Raw value
    :param max_length: Maximum string length (number of characters)
    :return: The output HTML code (input value is escaped)
    """
    val = str(value) if value else ""
    if len(val) <= max_length:
        return str(jinja2.escape(val))
    # else
    return '<span title="{}">{}</span>'.format(str(jinja2.escape(val)),
                                               str(jinja2.escape(val[:max_length])) + "...")


def pretty_field(value: str) -> str:
    """
    Format a snake_case field for display (snake_case -> Snake case)
    and make some other adjustments (identifier => id, remove meta_).
    :param value: Raw field name
    :return: Field to display
    """
    return value.replace("meta_", "").capitalize().replace("_", " ").replace("identifier", "id")



"""
Configuration and CLI management.
"""

import argparse
import configparser
import logging
from os.path import exists, realpath
from typing import Dict

from reportmix.config.property import ConfigProperty
from reportmix.errors import AppError
from reportmix.loaders import dependency_check, sonarqube, npm_audit
from reportmix.models import meta
from reportmix.models.issue import HASH_FIELDS

# Configuration global group name (for global configuration properties)
GLOBAL_CONFIG = "global"

# Global configuration properties
PROPERTIES = [
    ConfigProperty("output_dir", "the location to write the report", True, "./"),
    ConfigProperty("config_file", "the path to the configuration file", True, ".reportmix"),
    ConfigProperty("formats", "report formats to be generated (csv, html, json)",
                   True, "html", "^((F),)*(F)$".replace("F", "csv|html|json")),
    ConfigProperty("fields", "fields to include in the output report (CSV and HTML only)",
                   True, "all", r"^((\w+),)*(\w+)$"),
    ConfigProperty("hash", "fields to use for hash generation",
                   True, ",".join(HASH_FIELDS), r"^((\w+),)*(\w+)$"),
    ConfigProperty("title", "the HTML report title", True, "Issues Report", "^.{1,64}$"),
    ConfigProperty("logo", "the URL to the organization logo to display on the HTML report", False)
]


class ConfigBuilder:
    """
    Build the configuration from the configuration file and command-line arguments.
    """

    def __init__(self, version: str):
        """
        Initialize the configuration builder.
        :param version: Application version number
        """

        # Configuration properties
        self.properties = {
            GLOBAL_CONFIG: PROPERTIES,
            "meta": meta.PROPERTIES,
            "dependency_check": dependency_check.PROPERTIES,
            "npm_audit": npm_audit.PROPERTIES,
            "sonarqube": sonarqube.PROPERTIES
        }

        # Initialize the command-line argument parser
        self.parser = argparse.ArgumentParser(
            description='Merge reports from multiple tools into one single file.')
        self.parser.add_argument("-V", "--version", action="version",
                                 version="ReportMix " + version)
        self.parser.add_argument("-v", "--verbose", action="store_true",
                                 help="run verbosely (display DEBUG logging)")

        for group, props in self.properties.items():
            for prop in props:
                # Name
                if group == GLOBAL_CONFIG:
                    name = "--" + prop.name
                else:
                    name = "--" + group + "." + prop.name
                # Description
                description = prop.description
                if prop.default:
                    description += " (default: {})".format(prop.default)
                # Argument
                self.parser.add_argument(name, type=str, metavar=prop.name.upper(),
                                         default=argparse.SUPPRESS, help=description)

    def build(self) -> Dict[str, Dict[str, str]]:
        """
        Build configuration from CLI, file and default values.
        :return: Loaded configuration
        """

        # Build default configuration
        config = {}
        for group, props in self.properties.items():
            for prop in props:
                config.setdefault(group, {})[prop.name] = prop.default

        # Load configuration from command-line
        console_config = vars(self.parser.parse_args())

        # Configure logging
        logging_level = logging.DEBUG if console_config["verbose"] else logging.INFO
        logging.basicConfig(format='%(levelname)s\t| %(message)s', level=logging_level)

        # Load configuration from file
        logging.debug("Load configuration from file")
        if "config_file" in console_config:
            config_file_name = console_config["config_file"]
        else:
            config_file_name = config[GLOBAL_CONFIG]["config_file"]
        config_path = realpath(config_file_name)
        if exists(config_path):
            logging.debug("Load configuration from %s", config_path)
            parser = configparser.ConfigParser()
            parser.read(config_path)
            # Update configuration with file configuration values
            for group, props in self.properties.items():
                for prop in props:
                    if parser.has_option(group, prop.name):
                        config.setdefault(group, {})[prop.name] = parser[group][prop.name]

        # Append/override values from command-line
        logging.debug("Update configuration with command-line arguments")
        for group, props in self.properties.items():
            for prop in props:
                name = prop.name if group == GLOBAL_CONFIG else group + "." + prop.name
                if name in console_config:
                    config.setdefault(group, {})[prop.name] = console_config[name]

        # Check configuration properties
        err_count = 0
        for group, props in self.properties.items():
            for prop in props:
                name = prop.name if group == GLOBAL_CONFIG else group + "." + prop.name
                if prop.mandatory and not config[group][prop.name]:
                    logging.error("Property '%s' is required", name)
                    err_count += 1
                elif not prop.is_valid(config[group][prop.name]):
                    logging.error("Value of property '%s' is invalid", name)
                    err_count += 1
        if err_count > 0:
            logging.error("Configuration is incorrect, fix previous issues and run again")
            raise AppError()

        logging.debug("Configuration: %s", str(config))
        return config



"""
Main class.
"""

import logging
from os import path
from typing import Dict, Union, List

from reportmix.config.builder import GLOBAL_CONFIG
from reportmix.errors import AppError
from reportmix.exporters.csv import CsvExporter
from reportmix.exporters.html import HtmlExporter
from reportmix.exporters.json import JsonExporter
from reportmix.loaders.dependency_check import DependencyCheckLoader
from reportmix.loaders.npm_audit import NpmAuditLoader
from reportmix.loaders.sonarqube import SonarQubeLoader
from reportmix.models.issue import FLAT_FIELDS, Issue, HASH_FIELDS, select_fields
from reportmix.models.meta import Meta


class ReportMixer:
    """
    Merge reports from multiple tools into one single file.
    """

    def __init__(self, config: Dict[str, Union[str, Dict[str, str]]]):
        """
        Initialize the report mixer.
        :param config: Configuration.
        """
        self.config = config[GLOBAL_CONFIG]
        self.meta_config = config["meta"]
        self.loaders = {
            "dependency_check": DependencyCheckLoader(config["dependency_check"]),
            "npm_audit": NpmAuditLoader(config["npm_audit"]),
            "sonarqube": SonarQubeLoader(config["sonarqube"])
        }
        self.exporters = {
            "csv": CsvExporter(self.config),
            "json": JsonExporter(self.config),
            "html": HtmlExporter(self.config)
        }

    def merge(self) -> None:
        """
        Load and merge all available reports.
        """
        # Load and merge
        issues = self._load()
        if not issues:
            logging.warning("No issues to export, exiting")
            raise AppError()
        # Export
        self._export(issues)

    def _load(self) -> List[Issue]:
        """
        Load and merge issues from all loaders.
        Set metadata fields from configuration.
        :return: List of issues
        """
        # Load and merge
        issues = []
        logging.info("Merge reports: %s", ", ".join(self.loaders.keys()))
        for name, loader in self.loaders.items():
            logging.info("Loading %s report", name)
            issues.extend(loader.load())
        logging.info("Loaded %d issue(s)", len(issues))
        # Set metadata fields
        hash_fields = select_fields(self.config["hash"] or HASH_FIELDS)
        for issue in issues:
            issue.meta = Meta(self.meta_config["product"], self.meta_config["version"],
                              self.meta_config["organization"], self.meta_config["client"],
                              self.meta_config["audit_date"])
            issue.hash = issue.compute_hash(hash_fields)
        return issues

    def _export(self, issues: List[Issue]):
        """
        Export a list of issues to a report file.
        :param issues: Issues to export
        """
        # File
        output_dir: str = path.realpath(self.config["output_dir"])
        if not path.exists(output_dir) or not path.isdir(output_dir):
            logging.error("Invalid output directory %s", output_dir)
            raise AppError()

        # Fields (intersection between all fields and selected fields)
        only_fields = self.config["fields"].lower()
        fields = FLAT_FIELDS if only_fields == "all" else select_fields(only_fields)

        # Exporter
        for output_format in self.config["formats"].split(","):
            output_file_path = path.join(output_dir, "reportmix." + output_format)
            logging.debug("Exporting merged report (format: %s, fields: [%s])",
                          output_format, ", ".join(fields))
            self.exporters[output_format].export(output_file_path, issues, fields)
            logging.info("Merged report exported: %s", output_file_path)



"""
CSV report exporter.
"""

import csv
from typing import List

from reportmix.exporter import Exporter
from reportmix.models.issue import Issue


class CsvExporter(Exporter):
    """
    Export a merged report to a CSV file.
    """

    def export(self, output_file: str, issues: List[Issue], fields: List[str]):
        with open(output_file, "w", newline='', encoding='utf-8') as file:
            writer = csv.DictWriter(file, fieldnames=fields, extrasaction='ignore',
                                    delimiter=',', quotechar='"', quoting=csv.QUOTE_MINIMAL)
            writer.writeheader()
            for issue in issues:
                writer.writerow(issue.flatten())



"""
SonarQube report loader.
"""

import logging
from datetime import datetime
from typing import List

import requests

from reportmix.config.property import ConfigProperty
from reportmix.errors import AppError
from reportmix.loader import Loader
from reportmix.models.issue import Issue
from reportmix.models.project import Project
from reportmix.models.severity import SEVERITIES
from reportmix.models.subject import Subject
from reportmix.models.tool import Tool

# Possible values for types and statuses request parameters
TYPES = ["CODE_SMELL", "BUG", "VULNERABILITY", "SECURITY_HOTSPOT"]
DEFAULT_TYPES = TYPES[2]
STATUSES = ["OPEN", "CONFIRMED", "REOPENED", "RESOLVED", "CLOSED",
            "TO_REVIEW", "IN_REVIEW", "REVIEWED"]
DEFAULT_STATUSES = ",".join(STATUSES[0:3])

# Configuration properties
# https://docs.sonarqube.org/latest/analysis/analysis-parameters/
PROPERTIES: List[ConfigProperty] = [
    ConfigProperty("host_url", "the server URL", False, "http://localhost:9000"),
    ConfigProperty("login", "the login or authentication token"),
    ConfigProperty("password", "the password that goes with the login username"),
    ConfigProperty("project_key", "the project's unique key"),
    ConfigProperty("types", "issue types ({})".format(", ".join(TYPES)), False,
                   DEFAULT_TYPES, "^((T),)*(T)$".replace("T", "|".join(TYPES))),
    ConfigProperty("statuses", "issue statuses ({})".format(", ".join(STATUSES)), False,
                   DEFAULT_STATUSES, "^((S),)*(S)$".replace("S", "|".join(STATUSES)))
]


class SonarQubeLoader(Loader):
    """
    SonarQube project report loader using the Web API.
    """

    def load(self) -> List[Issue]:
        """
        Load project issues from the SonarQube Web API,
        parse them, map them, and return the list.
        :return: List of vulnerabilities.
        """
        cfg = self.config
        if not (cfg["host_url"] and cfg["project_key"]):
            logging.warning("SonarQube report ignored (required params: host_url, project_key)")
            return []

        # Authentication params
        auth = (cfg["login"] or "", cfg["password"] or "")

        # Fetch project info
        project_url = "{}/api/projects/search?q={}".format(cfg["host_url"], cfg["project_key"])
        try:
            resp = requests.get(project_url, auth=auth)
            project_resp = resp.json()["components"][0]
            project = Project(project_resp["key"], project_resp["name"], "")
        except Exception as ex:
            logging.error("Failed to get project information (%s)", ex)
            return []

        # Fetch issues info
        page_size = 500  # Number of issues in a result page
        page_index = 1  # Current page index
        total = page_size  # Total number of issues

        issues_base_url = "{}/api/issues/search?componentKeys={}&statuses={}&s=SEVERITY" \
                          "&asc=false&types={}&ps={}" \
            .format(cfg["host_url"], cfg["project_key"], DEFAULT_STATUSES,
                    cfg["types"] or DEFAULT_TYPES, page_size)
        try:
            issues = []
            # For each page (while total > max items fetched during the last iteration
            # and no more than 10000 issues have been requested)
            while total >= (page_index - 1) * page_size and page_index * page_size <= 10000:
                issues_url = issues_base_url + "&p=" + str(page_index)
                logging.debug("Fetching issues from %s", issues_url)
                # Request
                resp = requests.get(issues_url, auth=auth)
                result = resp.json()
                # Check response body
                if "paging" not in result or "issues" not in result:
                    logging.error("Server response is invalid ('paging' and 'issues' keys missing)")
                    raise AppError()
                total = result["paging"]["total"]
                fetched_count = (page_index - 1) * page_size + len(result["issues"])
                logging.debug("Fetched %d / %d issues", fetched_count, total)
                # Map issues
                for issue in result["issues"]:
                    # Severity
                    if issue["severity"] in SONARQUBE_SEVERITIES:
                        severity = SONARQUBE_SEVERITIES[issue["severity"]]
                    else:
                        severity = SEVERITIES[0]
                    # Subject location
                    location = issue["component"]
                    if "line" in issue:
                        location += ":" + str(issue["line"])
                    # Issue
                    issues.append(Issue(
                        ref=issue["key"],
                        identifier=issue["rule"],
                        name=issue["rule"],
                        type=issue["type"],
                        category=issue["type"],
                        description=issue["message"],
                        more=", ".join(issue["tags"]),
                        action=issue["message"],
                        effort=issue["effort"] if "effort" in issue else "",
                        analysis_date=datetime.strptime(project_resp["lastAnalysisDate"][:19],
                                                        "%Y-%m-%dT%H:%M:%S"),
                        severity=severity,
                        score="",
                        confidence="",
                        evidences=1,
                        source=issue["rule"],
                        source_date=datetime.strptime(issue["creationDate"][:19],
                                                      "%Y-%m-%dT%H:%M:%S"),
                        url="",
                        tool=Tool(
                            identifier="sonarqube",
                            name="SonarQube",
                            version=resp.headers["Sonar-Version"]
                        ),
                        subject=Subject(
                            identifier=issue["component"],
                            name=issue["component"],
                            description="",
                            version="",
                            location=location,
                            license=""
                        ),
                        project=project
                    ))
                page_index += 1  # Go to the next result page
            return issues
        except Exception as ex:
            logging.error("Failed to process issues (%s)", ex)
            return []


# SonarQube severities are a bit "excessive" so we define
# a custom map instead of using guess() function.
SONARQUBE_SEVERITIES = {
    "INFO": SEVERITIES[1],
    "MINOR": SEVERITIES[2],
    "MAJOR": SEVERITIES[3],
    "CRITICAL": SEVERITIES[4],
    "BLOCKER": SEVERITIES[5]
}



"""
Main issue model.
"""

import hashlib
import inspect
from datetime import datetime
from typing import Dict, Union, Optional, List

from reportmix.models.meta import Meta
from reportmix.models.project import Project
from reportmix.models.severity import Severity
from reportmix.models.subject import Subject
from reportmix.models.tool import Tool

# An issue as a dictionary with only first level items (no sub-dictionary or list).
FlatIssue = Dict[str, Union[str, int, datetime]]


class Issue:
    """
    An issue extracted from a report.
    """

    def __init__(self, ref: str, identifier: str, name: str, type: str, category: str,
                 description: str, more: str, action: str, effort: str,
                 analysis_date: Optional[datetime], severity: Severity, score: str,
                 confidence: str, evidences: int, source: str, source_date: Optional[datetime],
                 url: str, tool: Tool, subject: Subject, project: Project, meta: Meta = None,
                 hash: str = None):
        """
        Initialize an issue reported by a tool about a subject in a project.
        :param ref: Issue technical reference/identifier
        :param identifier: Issue unique identifier (CVE, rule, ...)
        :param name: Issue short name
        :param type: Issue type (UPPER_CASE)
        :param category: Issue category (e.g. CWE)
        :param description: A longer (but not too long) description of the issue
        :param more: More information about the issue
        :param action: Recommended action to solve the issue
        :param effort: Necessary effort to solve the issue (debt)
        :param analysis_date: Analysis / report creation date and time
        :param severity: Issue severity
        :param score: Computed score for this issue
        :param confidence: Confidence about this issue
        :param evidences: Number of evidences of this issue
        :param source: Issue source database
        :param source_date: Advisory, rule or ticket creation date
        :param url: Advisory, rule or ticket URL
        :param tool: Scan tool
        :param subject: Subject (application feature, file, dependency, etc.) affected by the issue
        :param project: Project affected by the issue
        :param meta: User-defined metadata
        :param hash: Issue stable unique generated identifier
        """
        self.ref = ref
        self.identifier = identifier
        self.name = name
        self.type = type
        self.category = category
        self.description = description
        self.more = more
        self.action = action
        self.effort = effort
        self.analysis_date = analysis_date
        self.severity = severity
        self.score = score
        self.confidence = confidence
        self.evidences = evidences
        self.source = source
        self.source_date = source_date
        self.url = url
        self.tool = tool
        self.subject = subject
        self.project = project
        self.meta = meta
        self.hash = hash

    def get_field(self, name: str) -> Optional[Union[str, int, datetime, Severity]]:
        """
        Get the value of a field by name.
        :param name: Name of the field from the FLAT_FIELDS list.
        :return: The value of the field
        """
        field = name.split("_", 1)
        # Sub-object
        if len(field) == 2 and field[0] in ["tool", "subject", "project", "meta"]:
            sub_obj = getattr(self, field[0])
            return getattr(sub_obj, field[1])
        # else: first-level field
        return getattr(self, name)

    def compute_hash(self, fields: List[str]) -> str:
        """
        Compute the issue unique identifier using the given fields.
        :param fields: Names of the fields to use in hash
        :return: The computed hash
        """
        if not fields:
            return ""
        raw_str = "".join([self.get_field(f) for f in fields])
        return hashlib.md5(str.encode(raw_str)).hexdigest()

    def to_dict(self) -> Dict:
        """
        Map the issue to a dictionary (deep mapping).
        :return: The issue as a dictionary
        """
        issue = vars(self).copy()
        issue["tool"] = vars(self.tool).copy()
        issue["subject"] = vars(self.subject).copy()
        issue["project"] = vars(self.project).copy()
        if self.meta:
            issue["meta"] = vars(self.meta).copy()
        return issue

    def flatten(self, sub_sep: str = "_") -> FlatIssue:
        """
        Flatten an issue to get a single level dictionary by mapping the object
        to a dictionary and bringing sub-dictionaries to the first level.
        Example: the attribute "name" of the sub-dictionary "project" will become
        an attribute of the root dictionary "project_name".
        :param sub_sep: Parent key and sub key separator
        :return: The flattened issue
        """
        dict_issue = self.to_dict()
        result = dict_issue.copy()
        for key in ["tool", "subject", "project", "meta"]:  # Sub-dictionaries
            if key in dict_issue and dict_issue[key]:
                for sub_key in dict_issue[key].keys():
                    result[key + sub_sep + sub_key] = dict_issue[key][sub_key]
                result.pop(key)
        return result


#
# Constants
#

# The list of issue fields
FIELDS = inspect.getfullargspec(Issue.__init__).args[1:]

# The list of issue fields after flattening
FLAT_FIELDS = [f for f in FIELDS if f not in ["tool", "subject", "project", "meta"]]
FLAT_FIELDS.extend(["tool_" + f for f in inspect.getfullargspec(Tool.__init__).args[1:]])
FLAT_FIELDS.extend(["subject_" + f for f in inspect.getfullargspec(Subject.__init__).args[1:]])
FLAT_FIELDS.extend(["project_" + f for f in inspect.getfullargspec(Project.__init__).args[1:]])
FLAT_FIELDS.extend(["meta_" + f for f in inspect.getfullargspec(Meta.__init__).args[1:]])

# Names of the default fields to use in hash generation
HASH_FIELDS = ["tool_identifier", "subject_identifier", "identifier"]


#
# Utilities
#

def select_fields(fields: Union[str, List[str]]) -> List[str]:
    """
    Select a list of field names from the full list of fields.
    :param fields: Field names list (as a list or a comma-separated string list)
    :return: Intersection between the list of issue fields and the fields argument
    """
    only_fields = fields
    if isinstance(fields, str):
        only_fields = map(lambda f: f.strip(), fields.split(","))
    return [f for f in only_fields if f in FLAT_FIELDS]



"""
Subject model.
"""


class Subject:
    """
    A subject affected by an issue (feature, file, class, dependency, ...).
    """

    def __init__(self, identifier: str, name: str, description: str, version: str,
                 location: str, license: str):
        """
        Initialize a subject affected by an issue.
        :param identifier: Subject unique identifier (file path, fully qualified name, hash, ...)
        :param name: Subject short name (class name, dependency name, ...)
        :param description: A longer (but not too long) description of the subject
        :param version: Version(s) of the subject affected by the issue
        :param location: The location of the subject (full file path, full package name, URL, ...)
        :param license: Subject license
        """
        self.identifier = identifier
        self.name = name
        self.description = description
        self.version = version
        self.location = location
        self.license = license



"""
Project model.
"""


class Project:
    """
    A scanned project.
    """

    def __init__(self, identifier: str, name: str, version: str):
        """
        Initialize a scanned project.
        :param identifier: Project unique identifier
        :param name: Project short common name
        :param version: Project version (semantic version or something else)
        """
        self.identifier = identifier
        self.name = name
        self.version = version



"""
Report loader parent class.
"""

from typing import List, Dict

from reportmix.models.issue import Issue


class Loader:
    """
    A report loader.
    Load and parse a specific type of report associated to a tool.
    """

    def __init__(self, config: Dict[str, str]):
        """
        Initialize the report loader with the given configuration.
        :param config: Report loader configuration.
        """
        self.config = config

    def load(self) -> List[Issue]:
        """
        Load the report and return the list of items.
        :return: List of items loaded from the report.
        """
        return []



"""
Configuration property.
"""

import re


class ConfigProperty:
    """
    A configuration property available through the CLI or file configuration.
    """

    def __init__(self, name: str, description: str, mandatory: bool = False,
                 default: str = None, pattern: str = None):
        """
        Initialize a configuration property.
        :param name: Property name (format: snake_case)
        :param description: Property short help message
        :param mandatory: Set the property as required (a loader property can't be mandatory,
        if a property is missing, log a warning and ignore this loader)
        :param default: Property default value
        :param pattern: Validation pattern
        """
        self.name = name
        self.description = description
        self.mandatory = mandatory
        self.default = default
        self.pattern = pattern

    def is_valid(self, value) -> bool:
        """
        Check if a given value is valid for this property.
        :param value: The value to test
        :return: true if the value is valid
        """
        required_ok = not self.mandatory or value
        pattern_ok = not self.pattern or re.match(self.pattern, value)
        return required_ok and pattern_ok



"""
Tool model.
"""


class Tool:
    """
    A tool that scanned a project and found issues.
    """

    def __init__(self, identifier: str, name: str, version: str):
        """
        Initialize a tool that scanned a project and found issues.
        :param identifier: Scan tool id
        :param name: Scan tool display name
        :param version: Scan tool version
        """
        self.identifier = identifier
        self.name = name
        self.version = version



"""
Metadata model.
"""

from datetime import datetime
from typing import List

from reportmix.config.property import ConfigProperty


class Meta:
    """
    Metadata (user-defined global fields).
    """

    def __init__(self, product: str, version: str, organization: str, client: str,
                 audit_date: str):
        """
        Initialize metadata.
        :param product: Product name
        :param version: Product version
        :param organization: Organization name
        :param client: Client name
        :param audit_date: Audit date (default: now())
        """
        self.product = product
        self.version = version
        self.organization = organization
        self.client = client
        self.audit_date = audit_date if audit_date and audit_date != "now()" \
            else str(datetime.now().replace(microsecond=0))


# Metadata configuration properties
PROPERTIES: List[ConfigProperty] = [
    ConfigProperty("product", "the product name"),
    ConfigProperty("version", "the product version"),
    ConfigProperty("organization", "the organization name"),
    ConfigProperty("client", "the client name"),
    ConfigProperty("audit_date", "the audit date", False, "now()")
]



"""
JSON report exporter.
"""

import json
from typing import List

from reportmix.exporter import Exporter
from reportmix.models.issue import Issue


class JsonExporter(Exporter):
    """
    Export a merged report to a JSON file.
    """

    def export(self, output_file: str, issues: List[Issue], fields: List[str]):
        with open(output_file, "w") as file:
            json.dump([i.to_dict() for i in issues], file, default=str)



"""
ReportMix setup file.
"""

import re
from io import open
from os import path

from setuptools import setup, find_packages

DIR = path.abspath(path.dirname(__file__))


def get_long_description() -> str:
    """
    Get the long description from the README.md file.
    :return: Long description
    """
    with open(path.join(DIR, 'README.md'), encoding='utf-8') as file:
        return file.read()


def get_version() -> str:
    """
    Get the version number from reportmix/main.py.
    :return: Version number
    """
    main_file = path.join(DIR, 'reportmix', 'main.py')
    version_re = re.compile(r'__version__\s+=\s+"(?P<version>.*)"')
    with open(main_file, 'r', encoding='utf-8') as file:
        match = version_re.search(file.read())
        return match.group('version') if match is not None else 'unknown'


setup(
    name='reportmix',
    version=get_version(),
    description='Merge reports from multiple tools into a single file',
    long_description=get_long_description(),
    long_description_content_type='text/markdown',
    url='https://github.com/GaelGirodon/reportmix',
    author='Gael Girodon',
    author_email='contact@gaelgirodon.fr',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Quality Assurance',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7'
    ],
    keywords='report mix merge security dependency-check npm audit sonarqube owasp',
    packages=find_packages(exclude=['tests*']),
    include_package_data=True,
    python_requires='>=3.7',
    install_requires=[
        'requests>=2.22.0',
        'jinja2>=2.10.1'
    ],
    entry_points={
        'console_scripts': [
            'reportmix=reportmix.main:main'
        ]
    },
    project_urls={
        'Bug Reports': 'https://github.com/GaelGirodon/reportmix/issues',
        'Source': 'https://github.com/GaelGirodon/reportmix',
    },
)



"""
Report exporter parent class.
"""

from typing import List, Dict

from reportmix.models.issue import Issue


class Exporter:
    """
    A merged report exporter.
    Export a list of issues to a file.
    """

    def __init__(self, config: Dict[str, str]):
        """
        Initialize the report exporter with the given configuration.
        :param config: Report exporter configuration.
        """
        self.config = config

    def export(self, output_file: str, issues: List[Issue], fields: List[str]):
        """
        Export a list of issues to a file.
        :param output_file: Path to the output file.
        :param issues: List of issues to write.
        :param fields: List of fields to include in the output report.
        """



#!/usr/bin/env python

"""
Application main function and version.
"""

import sys

from reportmix.config.builder import ConfigBuilder
from reportmix.errors import AppError
from reportmix.mixer import ReportMixer

__version__ = "0.4.1"


def main():
    """
    Entry point for the application script.
    """

    # Load configuration
    try:
        config = ConfigBuilder(__version__).build()
    except AppError:
        sys.exit(1)

    # Merge reports
    try:
        ReportMixer(config).merge()
    except AppError:
        sys.exit(2)


if __name__ == "__main__":
    main()



"""
npm audit report loader.
"""

import json
import logging
from datetime import datetime
from os import path
from typing import List

from reportmix.config.property import ConfigProperty
from reportmix.loader import Loader
from reportmix.models import severity
from reportmix.models.issue import Issue
from reportmix.models.project import Project
from reportmix.models.subject import Subject
from reportmix.models.tool import Tool

# Configuration properties
PROPERTIES: List[ConfigProperty] = [
    ConfigProperty("report_file", "path to the report file", True, "npm-audit.json")
]


class NpmAuditLoader(Loader):
    """
    npm security audit report loader.
    """

    def load(self) -> List[Issue]:
        """
        Load the npm audit report in JSON format,
        parse it, map vulnerabilities to issues, and return the list.
        :return: List of vulnerabilities.
        """
        report_file_path = path.realpath(self.config["report_file"])
        if not (report_file_path.endswith(".json") and path.exists(report_file_path)):
            logging.warning("npm audit report ignored (file not found or not *.json)")
            return []
        logging.debug("Loading report %s", report_file_path)
        try:
            with open(report_file_path, "r", encoding="utf8") as report_file:
                report = json.load(report_file)
                advisories = report["advisories"]
                issues = []
                for number, adv in advisories.items():
                    for finding in adv["findings"]:
                        issues.append(Issue(
                            ref=adv["id"] or number,
                            identifier=", ".join(adv["cves"]) or adv["title"],
                            name=adv["title"],
                            type="VULNERABILITY",
                            category=adv["cwe"],
                            description=str(adv["overview"]).strip(),
                            more="",
                            action=str(adv["recommendation"]).strip(),
                            effort="",
                            analysis_date=None,
                            severity=severity.guess(adv["severity"]),
                            score="",
                            confidence="",
                            evidences=len(adv["findings"]),
                            source="NPM Public Advisories",
                            source_date=datetime.strptime(adv["created"][:19], "%Y-%m-%dT%H:%M:%S"),
                            url=adv["url"],
                            tool=Tool(
                                identifier="npm_audit",
                                name="npm audit",
                                version=""
                            ),
                            subject=Subject(
                                identifier=adv["module_name"],
                                name=adv["module_name"],
                                description="",
                                version="Vulnerable versions: {}, Patched versions: {}".format(
                                    adv["vulnerable_versions"], adv["patched_versions"]),
                                location=finding["paths"][0] if finding["paths"] else "",
                                license=""
                            ),
                            project=Project(
                                identifier="",
                                name="",
                                version=""
                            )
                        ))
                return issues
        except Exception as ex:
            logging.error("Failed to load, parse and map the report (%s)", ex)
            return []
